/**
  *  \file main.cpp
  *  \brief Little Ugly Test Program
  */

#include <iostream>
#include <fstream>
#include <cstdlib>
#include "data.h"
#include "game.h"
#include "util.h"
#include "osdep.h"
#include "program.h"
#include "action.h"
#include "race.h"
#include "expander.h"
#include "selection.h"
#include "loaderstate.h"
#include "net/http.h"
#include "net/smtp.h"
#include "net/url.h"
#include "net/mailtemplate.h"

static bool
askYesNo(std::string prompt)
{
    std::string reply;
    std::cout << prompt << " (y/n)? ";
    return getline(std::cin, reply) && reply == "y";
}

class TurnUploader : public ValueEnumerator<Race> {
    Game& m_Game;
 public:
    TurnUploader(Game& g)
        : m_Game(g)
        { }

    void handleItem(Race&);

    void doUpload(Url& u, Race& r);

    void doMail(const MailTemplate& t, Race& r);
};

class ResultDownloader : public ValueEnumerator<Race> {
    Game& m_Game;
    Action& m_Action;
 public:
    ResultDownloader(Game& g, Action& a)
        : m_Game(g), m_Action(a)
        { }

    void handleItem(Race&);
};

void
TurnUploader::doUpload(Url& u, Race& r)
{
    // upload race's turn file
    std::string basename = "player" + itoa(r.getRaceNumber()) + ".trn";
    std::string tname = makeFileName(m_Game.getLocalDirectory(), basename);
    std::ifstream trn(tname.c_str(), std::ifstream::binary);
    if (!trn) {
        std::cout << ">> " << tname << " does not exist, run maketurn\n";
        return;
    }

    // build packet to upload
    MimeBuilder builder("multipart/form-data");
    MimeParameterEncoder mpe(builder);
    u.trimParameters(mpe);

    // Now add the file
    builder.addBoundary();
    builder.addHeader("Content-Disposition",
                      std::string("form-data; name=\"") + r.getUploadId() + "\"; filename=\"" + basename + "\"");
    builder.addHeader("Content-Type", "application/octet-stream");
    builder.addHeader("Content-Transfer-Encoding", "base64");
    char buf[4096];
    int count;
    while ((count = trn.rdbuf()->sgetn(buf, sizeof(buf))) > 0) {
        builder.addBase64(buf, count);
    }
    builder.addBoundary();
    builder.finish();

    // Upload
    std::cout << "Now posting to " << u.toString() << "...\n";

    Http req;
    if (req.post(u, builder, std::cout)) {
        if (req.isSuccessfulReply()) {
            std::cout << "OK\n";
        } else {
            std::cout << "Not OK: " << req.getReplyLine() << "\n";
        }
    } else {
        std::cout << "Failure: " << req.getErrorMessage() << ", " << req.getServerErrorMessage() << "\n";
    }
}

void
TurnUploader::doMail(const MailTemplate& t, Race& r)
{
    std::string from_addr = Smtp::getFromAddress();

    // mail race's turn file
    std::string basename = "player" + itoa(r.getRaceNumber()) + ".trn";
    std::string tname = makeFileName(m_Game.getLocalDirectory(), basename);
    std::ifstream trn(tname.c_str(), std::ifstream::binary);
    if (!trn) {
        std::cout << ">> " << tname << " does not exist, run maketurn\n";
        return;
    }

    // build packet to upload
    MimeBuilder mb("multipart/mixed");
    mb.addHeader("From", from_addr);
    mb.addHeader("To", t.getAddress());
    if (t.getSubject().size())
        mb.addHeader("Subject", t.getSubject());
    else
        mb.addHeader("Subject", "Turn file for " + m_Game.getName());
    mb.addHeader("X-Mailer", "VGAPIntegrator/0.5");
    mb.addHeader("MIME-Version", "1.0");

    /* Text part */
    mb.addBoundary();
    mb.addHeader("Content-Type", "text/plain; charset=iso-8859-1");
    mb.addHeader("Content-Disposition", "inline");
    mb.addHeader("Content-Transfer-Encoding", "quoted-printable");
    if (t.getBody().size()) {
        mb.addLineQP(t.getBody());
    } else {
        mb.addLineQP("Please find attached the turn file for '" + m_Game.getName() + "' player " +
                     itoa(r.getRaceNumber()) + ".");
    }
    mb.addLineQP("");

    /* File part */
    mb.addBoundary();
    mb.addHeader("Content-Type", "application/octet-stream");
    mb.addHeader("Content-Disposition", "attachment; filename=\"" + basename + "\"");
    mb.addHeader("Content-Transfer-Encoding", "base64");
    char buf[4096];
    int count;
    while ((count = trn.rdbuf()->sgetn(buf, sizeof(buf))) > 0) {
        mb.addBase64(buf, count);
    }
    mb.addBoundary();
    mb.finish();

    std::cout << "Now mailing...\n";

    Smtp mailer;
    if (mailer.sendMail(from_addr, t.getAddress(), mb))
        std::cout << "OK\n";
    else
        std::cout << "Failure: " << mailer.getErrorMessage() << ", " << mailer.getServerErrorMessage() << "\n";
}

void
TurnUploader::handleItem(Race& r)
{
    // figure out URL
    {
        Url u;
        if (u.parse(r.getUploadUrl())) {
            if (u.getProtocol() != "http" && u.getProtocol() != "http-post") {
                std::cout << ">> Only http upload implemented so far\n";
                return;
            }
            doUpload(u, r);
            return;
        }
    }

    // or do we want to mail?
    {
        MailTemplate t;
        if (t.parse(r.getUploadUrl())) {
            doMail(t, r);
            return;
        }
    }

    std::cout << ">> Cannot handle upload address '" << r.getUploadUrl() << "'\n";
}

static void
doDownloadFile(std::string url, Game& g)
{
    Url u;
    if (!u.parse(url)) {
        std::cout << ">> invalid URL\n\n";
        return;
    }

    if (u.getProtocol() != "http" && u.getProtocol() != "http-post") {
        std::cout << ">> unsupported download protocol\n";
        return;
    }

    std::string tmpname = makeFileName(g.getLocalDirectory(), "__vgapi.tmp");
    std::ofstream out(tmpname.c_str(), std::ofstream::binary);

    std::cout << "Downloading '" << u.toString() << "'\n";

    Http downloader;
    bool result;
    if (u.getProtocol() == "http") {
        // GET
        result = downloader.downloadFile(u, out);
    } else {
        // POST
        MimeBuilder mb("multipart/form-data");
        MimeBuilder builder("multipart/form-data");
        MimeParameterEncoder mpe(mb);
        u.trimParameters(mpe);
        mb.addBoundary();
        mb.finish();
        result = downloader.post(u, mb, out);
    }
    out.close();

    if (!result) {
        std::cout << "Failure:, " << downloader.getErrorMessage() << ", " << downloader.getServerErrorMessage()
                  << "\n";
        std::remove(tmpname.c_str());
    } else if (!downloader.isSuccessfulReply()) {
        std::cout << "Not OK: " << downloader.getReplyLine() << "\n";
        std::remove(tmpname.c_str());
    } else {
        // Figure out the name of the file
        std::string outname;
        outname = getParameter(downloader.getHeader("Content-Disposition"), "filename");
        if (outname.size() == 0)
            outname = getParameter(downloader.getHeader("Content-Type"), "filename");
        if (outname.size() == 0)
            outname = u.getBaseName();

        outname = getBaseName(outname);

        std::cout << "OK, received file " << outname << "\n";

        // uh-oh
        std::remove(outname.c_str());
        std::rename(tmpname.c_str(), makeFileName(g.getLocalDirectory(), outname).c_str());
    }
}

void
ResultDownloader::handleItem(Race& r)
{
    Selection sel;
    sel.selectGame(&m_Game);
    sel.selectRace(&r);
    sel.selectAction(&m_Action);

    std::string url = sel.expandString(r.getDownloadUrl());
    if (url.size() == 0) {
        std::cout << ">> don't know how to download result for " << r.getRaceNumber() << "\n";
    } else {
        doDownloadFile(url, m_Game);
    }
}

static void
runAction(Action& a, Selection& sel)
{
    sel.selectAction(&a);
    std::string s = sel.expandString(a.getCommandLine());
    if (s.size() == 0) {
        std::cout << ">> Command not possible\n";
    } else {
        std::string verb = eatWord(s);
        std::cout << "\n>> Verb '" << verb << "'"
                  << "\n>> Args '" << s << "'\n\n";
        if (verb == "command") {
            while (s.size() && s[0] == '-')
                eatWord(s);
            std::cout << ">> Would now run '" << s << "'\n\n";
        } else if (verb == "browser") {
            if (s.size()) {
                system(("lynx '" + s + "'").c_str());
            } else {
                std::cout << "Missing url.\n\n";
            }
        } else if (verb == "builtin-download-rst") {
            if (!sel.getSelectedGame()) {
                std::cout << ">> no game\n\n";
            } else {
                ResultDownloader rd(*sel.getSelectedGame(), a);
                sel.getSelectedGame()->enumerateRaces(rd);
            }
        } else if (verb == "builtin-download-spec") {
            if (!s.size()) {
                std::cout << ">> missing argument to 'builtin-download-spec'\n\n";
            } else if (!sel.getSelectedGame()) {
                std::cout << ">> no game\n\n";
            } else {
                doDownloadFile(s, *sel.getSelectedGame());
            }
        } else if (verb == "builtin-upload-trn") {
            if (sel.getSelectedGame()) {
                TurnUploader tu(*sel.getSelectedGame());
                sel.getSelectedGame()->enumerateRaces(tu);
            }
        } else {
            std::cout << "Unknown verb.\n\n";
        }
    }
}

class GameDirProvider : public ValueEnumerator<Game> {
    Data& d;
 public:
    GameDirProvider(Data& d)
        : d(d)
        { }
    void handleItem(Game& g) {
        if (g.getLocalDirectory().empty()) {
            std::string n = d.getNewGameDirectory(g.getName(), false);
            if (n.empty()) {
                std::cout << "I cannot allocate a Winplan slot for " << g.getName() << "\n"
                          << "You can assign a directory outside the Winplan tree, which is\n"
                          << "then only accessible to non-Winplan programs.\n"
                          << "Or, leave this directory unassigned, make room by deleting other\n"
                          << "games, and invoke the 'assign' function later to assign this game\n"
                          << "its Winplan slot.\n";
                if (!askYesNo("Assign outside directory"))
                    return;
                n = d.getNewGameDirectory(g.getName(), true);
            }
            g.setLocalDirectory(n);
        }
    }
};

static void
gameMenu(Data& d, Game& g)
{
    std::vector<Program*> p;
    VectorValueEnumerator<Program> pi(p);
    d.enumeratePrograms(pi);

    std::vector<Action*> a;
    VectorValueEnumerator<Action> ai(a);
    d.enumerateSimpleActions(g, ai);

    std::vector<Race*> r;
    VectorValueEnumerator<Race> ri(r);
    g.enumerateRaces(ri);

    Selection sel;
    sel.selectGame(&g);

    std::cout << "-- Game " << g.getName() << " --\n";
    if (sel.getSelectedRace())
        std::cout << "r: race (current is " << sel.getSelectedRace()->getRaceNumber() << ")\n";
    else
        std::cout << "r: race (none selected)\n";
    for (unsigned int i = 0; i < a.size(); ++i)
        std::cout << i << ": " << a[i]->getDescription() << " ("
                  << a[i]->getName() << ", using "
                  << a[i]->getProgram().getName() << ")\n";
    for (unsigned int i = 0; i < p.size(); ++i)
        std::cout << "p" << i << ": " << p[i]->getName() << " ->\n";
    while (1) {
        std::string reply;
        std::cout << "Main Game> ";
        if (!getline(std::cin, reply))
            break;
        int n;
        if (reply == "q") {
            break;
        } else if (reply == "r") {
            for (unsigned int i = 0; i < r.size(); ++i)
                std::cout << i << ": race " << r[i]->getRaceNumber() << (r[i] == sel.getSelectedRace() ? " <-\n" : "\n");
            if (r.size() == 0)
                std::cout << "No races defined for this game\n";
            std::cout << "Main Game Race> ";
            if (getline(std::cin, reply) && atoi(reply, n) && n >= 0 && n < int(r.size()))
                sel.selectRace(r[n]);
        } else if (atoi(reply, n) && n >= 0 && n < int(a.size())) {
            runAction(*a[n], sel);
        } else if (reply.size() && reply[0] == 'p' && atoi(reply.substr(1), n) && n >= 0 && n < int(p.size())) {
            std::vector<Action*> pa;
            VectorValueEnumerator<Action> pai(pa);
            p[n]->enumerateActions(pai);
            std::cout << "::: Program " << p[n]->getName() << " :::\n";
            for (unsigned int i = 0; i < pa.size(); ++i)
                std::cout << i << ": " << pa[i]->getDescription() << "\n";
            std::cout << "Main Game Program> ";
            if (getline(std::cin, reply) && atoi(reply, n) && n >= 0 && n < int(pa.size())) {
                runAction(*pa[n], sel);
            }
        } else {
            std::cout << "huh?\n";
        }
    }
}

static void
loadFile(Data& d)
{
    std::string s, reply;
    std::cout << "Filename?";
    if (getline(std::cin, s)) {
        std::ifstream in(s.c_str());
        if (!in) {
            std::cerr << "can't open\n";
        } else {
            try {
                Data tmp;
                LoaderState st;
                st.setBaseDirectory(getDirectoryName(s));
                tmp.read(in, st);
                std::cout << "File contains " << st.getNumItems(st.Globals) << " global config definitions,\n"
                          << "  " << st.getNumItems(st.Programs) << " program definitions,\n"
                          << "  " << st.getNumItems(st.Games) << " game definitions.\n";

                st.setAccept(st.Globals,  st.getNumItems(st.Globals)  != 0 && askYesNo("Read globals"));
                st.setAccept(st.Programs, st.getNumItems(st.Programs) != 0 && askYesNo("Read programs"));
                st.setAccept(st.Games,    st.getNumItems(st.Games)    != 0 && askYesNo("Read games"));
                if (st.isAcceptAny()) {
                    in.clear(); // weil wir am Ende waren
                    in.seekg(0);
                    d.read(in, st);

                    GameDirProvider gdp(d);
                    d.enumerateGames(gdp);
                }
            }
            catch (std::exception& e) {
                std::cout << "Load aborted: " << e.what() << "\n";
            }
        }
    }
}

static void
newGame(Data& d)
{
    std::string reply;
    std::string name;
    std::cout << "Game name?";
    if (!getline(std::cin, name))
        return;

    if (d.getGameByName(name)) {
        std::cout << "haben wir schon.\n";
        return;
    }

    std::string dir = d.getNewGameDirectory(name, false);
    if (!dir.size()) {
        std::cout << "You have Winplan, and I cannot find a free game slot.\n";
        if (!askYesNo("Continue and make the game outside the Winplan tree?"))
            return;
        dir = d.getNewGameDirectory(name, true);
    }
    std::cout << "Game will be in '" << dir << "'\n";

    Game* g = new Game(name);
    d.addGame(g);
    g->setLocalDirectory(dir);

    // Noch eine Rasse anlegen
    int nr;
    std::cout << "What race do you play?";
    if (getline(std::cin, reply) && atoi(reply, nr) && nr > 0 && nr <= 11) {
        g->addRace(new Race(nr));
    }
}

static void
netConfig()
{
    Url u;
    std::string reply;
    std::string reply2;
    int port;

    // proxy
    do {
        std::cout << "Proxy (just ENTER for none)?";
        if (!getline(std::cin, reply))
            return;
    } while (reply.size() && !u.parse(reply));
    Http::setProxy(u);

    // mail address
    std::cout << "Mail address?";
    getline(std::cin, reply);
    Smtp::setFromAddress(reply);

    // SMTP host
    std::cout << "SMTP server?";
    getline(std::cin, reply);
    std::cout << "SMTP port (ENTER for default)?";
    getline(std::cin, reply2);
    port = -1;
    atoi(reply2, port);
    Smtp::setServer(reply, port);
}

static void
mainMenu(Data& d)
{
    std::string reply;
    while (1) {
        std::cout << "=== Game List ===\n";
        std::vector<Game*> g;
        {
            VectorValueEnumerator<Game> vi(g);
            d.enumerateGames(vi);
        }
        std::cout << "load: load definition\n"
                  << "new: new game\n"
                  << "assign: assign missing local dirs\n"
                  << "q: quit\n"
                  << "print: print config file to stdout\n"
                  << "nc: network config\n"
                  << "rm: remove game\n";
        for (std::vector<Game*>::size_type i = 0; i < g.size(); ++i)
            std::cout << i << ": " << g[i]->getName() << "\n";
        std::cout << "Main> ";
        if (!getline(std::cin, reply))
            return;
        if (reply == "print") {
            d.write(std::cout);
        } else if (reply == "load") {
            loadFile(d);
        } else if (reply == "q") {
            break;
        } else if (reply == "assign") {
            GameDirProvider gdp(d);
            d.enumerateGames(gdp);
        } else if (reply == "rm") {
            int nr;
            std::cout << "Which one?";
            if (getline(std::cin, reply) && atoi(reply, nr) && nr >= 0 && nr < int(g.size()))
                d.deleteGame(g[nr]);
        } else if (reply == "nc") {
            netConfig();
        } else if (reply == "new") {
            newGame(d);
        } else {
            int num;
            if (atoi(reply, num) && num >= 0 && num < int(g.size()))
                gameMenu(d, *g[num]);
            else
                std::cout << "h?\n";
        }
    }
}

int main(int argc, char** argv)
{
    Data d;
    std::string filename = getSettingsFileName();
    if (argc == 2) {
        filename = argv[1];
    }

    {
        std::ifstream in(filename.c_str());
        if (in) {
            LoaderState st;
            st.setBaseDirectory(getDirectoryName(filename));
            d.read(in, st);
        }
    }
    mainMenu(d);

    {
        std::string bak = filename + ".bak";
        std::remove(bak.c_str());
        std::rename(filename.c_str(), bak.c_str());
        std::ofstream out(filename.c_str());
        if (out) {
            d.write(out);
        }
    }
}
