/**
  *  \file os/nix/socket.h
  *  \brief Socket class for Unixoid systems / BSD sockets
  *  \author Stefan Reuther
  */
#ifndef SOCKET_H_INCLUDED
#define SOCKET_H_INCLUDED

#include <string>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>

/** Lightweight socket class. This class is a replacement for the
    OS-dependant type of a socket handle (e.g. int on Unix, SOCKET on
    Windows). It is used almost the same as the OS-dependant socket
    handle is used. In particular, this has no sharing / RAII
    semantics or exceptions. */
class Socket {
    int m_Socket;
 public:
    /** Default constructor. \post !isValid(). */
    Socket()
        : m_Socket(-1)
        { }

    /** Create a socket. Equivalent to socket(2). Use isValid() to
        check whether the operation worked. */
    Socket(int af, int type, int proto)
        : m_Socket(::socket(af, type, proto))
        { }

    /** Check whether this is a valid socket. */
    bool isValid()
        { return m_Socket >= 0; }

    /** Connect to specified address. Equivalent of connect(2).
        \return true on success */
    bool connect(const struct sockaddr *serv_addr, socklen_t addrlen)
        { return ::connect(m_Socket, serv_addr, addrlen) == 0; }

    /** Write data to socket. Equivalent of write(2).
        \return amount written */
    int write(const void* buf, unsigned int size)
        {
            errno = 0;
            return ::write(m_Socket, buf, size);
        }

    /** Read data from socket. Equivalent of read(2).
        \return amount read */
    int read(void* buf, unsigned int size)
        { return ::read(m_Socket, buf, size); }

    /** Close socket. Equivalent of close(2). \post !isValid */
    int close()
        {
            int rv = ::close(m_Socket);
            m_Socket = -1;
            return rv;
        }

    /** Get last error message. */
    static std::string getError()
        { return strerror(errno); }

    static std::string getResolverError()
        { return hstrerror(h_errno); }
};

#endif
