/**
  *  \file url.cpp
  *  \brief URL parsing
  *
  *  \see RFC 1630 Universal Resource Identifiers
  */

#include "net/url.h"
#include "util.h"
#include "encoding.h"

/** Constructor. */
Url::Url()
    : m_Port(0), m_Valid(false)
{ }

/** Destructor. */
Url::~Url()
{ }

/** Parse an URL. If the URL is syntactically valid, store it in this
    object. Otherwise, does not modify it.
    \param s URL string
    \return true on success */
bool
Url::parse(const std::string s)
{
    std::string::size_type n = s.find(':');
    if (n == s.npos)
        /* No protocol identifier */
        return false;
    if (s.substr(n, 3) != "://")
        /* No "://" */
        return false;

    /* ok, found a protocol */
    std::string proto(s, 0, n);
    n += 3;

    std::string::size_type hostend = s.find_first_of(":/?#", n);
    std::string::size_type portend = hostend;
    int port = -1;
    if (hostend != s.npos && s[hostend] == ':') {
        ++portend;
        port = 0;
        while (portend < s.size() && s[portend] >= '0' && s[portend] <= '9') {
            port = 10*port + s[portend] - '0';
            portend++;
        }
        if (portend == hostend+1)
            port = -1;
    }

    /* now, portend must point to the end of the string or at the '/'
       that begins the file name */
    if (portend < s.size()) {
        if (s[portend] == '?' || s[portend] == '#') {
            /* http://localhost?qry */
            m_Path.assign("/");
            m_Path.append(s, portend, s.npos);
        } else if (s[portend] == '/') {
            m_Path.assign(s, portend, s.npos);
        } else {
            return false;
        }
    } else {
        m_Path.assign("/");
    }
    m_Protocol = proto;
    m_Host.assign(s, n, hostend-n);
    m_Port = port;
    m_Valid = true;
    return true;
}

/** Convert content of this object back into string. Usually returns
    the same string as given to parse(). */
std::string
Url::toString() const
{
    if (!m_Valid)
        return std::string();

    std::string port;
    if (m_Port >= 0)
        port = ":" + itoa(m_Port);
    
    return m_Protocol + "://" + m_Host + port + m_Path;
}

/** Get base name of URL. For http://host/dir/file, returns file.
    Returns "index.html" if URL contains no file part. */
std::string
Url::getBaseName() const
{
    if (!m_Valid)
        return std::string();

    std::string path = m_Path;
    std::string::size_type n = path.find_first_of("?#");
    if (n != path.npos)
        path.erase(n);

    n = path.rfind('/');
    if (n == std::string::npos || n+1 == path.size())
        return std::string("index.html");
    else
        return std::string(path, n+1);
}

/** Remove parameters from this URL. If this URL contains parameters
    (query), remove those and add them to \c e. */
void
Url::trimParameters(KVEnumerator& e)
{
    trimParameters(m_Path, e);
}

/** Remove parameters from string. If this string contains a '?',
    removes the query string after it and assigns the query content to
    \c e. */
void
Url::trimParameters(std::string& s, KVEnumerator& e)
{
    std::string::size_type beg = s.find('?');
    if (beg == s.npos)
        return;

    /* we have parameters */
    std::string::size_type n = beg;
    while (n != s.npos) {
        std::string::size_type end = s.find('&', n+1);
        std::string here(s, n+1, end == s.npos ? s.npos : end - n - 1);
        std::string::size_type eq = here.find('=');
        if (eq != s.npos) {
            e.handleItem(decodeUrl(here.substr(0, eq)),
                         std::string(),
                         decodeUrl(here.substr(eq+1)));
        }
        n = end;
    }
    s.erase(beg);
}
